/*
--          This file is part of the New World OS and Objectify projects
--         Copyright (C) 2005, 2006, 2007, 2008, 2009, 2011  QRW Software
--               J. Scott Edwards - j.scott.edwards.nwos@gmail.com 
--
--   This program is free software: you can redistribute it and/or modify
--   it under the terms of the GNU General Public License as published by
--   the Free Software Foundation, either version 3 of the License, or
--   (at your option) any later version.
--
--   This program is distributed in the hope that it will be useful,
--   but WITHOUT ANY WARRANTY; without even the implied warranty of
--   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--   GNU General Public License for more details.
--
--   You should have received a copy of the GNU General Public License
--   along with this program, in the file LICENSE.  If not, see 
--   <http://www.gnu.org/licenses/>.
--
--   For the latest information, source code (SVN), releases, and bug tracking
--   go to:
--      http://savannah.nongnu.org/projects/objectify
--
--   For releases from Alpha_30 and up, bug and feature request tracking go to:
--      http://sourceforge.net/projects/objectify
--
--   For older bug tracking, releases and source code (CVS) prior to the
--   Alpha_30 release go to:
--      http://sourceforge.net/projects/nwos
--
--   Other related websites:
--      http://www.qrwsoftware.com
--      http://www.worldwide-database.org
--
--   You can also contact me via paper mail at:
--
--      QRW Software
--      P.O. Box 27511
--      Salt Lake City, UT 84127-0511, USA.
--
--   $Author: jsedwards $
--   $Date: 2011-11-28 14:40:21 -0700 (Mon, 28 Nov 2011) $
--   $Revision: 4970 $
--
--   NOTE: Subversion does not support the Log keyword so I have removed the
--   logs that were here when I was using CVS.  Use the "svn log" command to
--   see the revisions of this file.  Since this file was originally created
--   from a file called "create_person.c", I left the part of the log from
--   when it was called create_person.c below to make it easier to access.
--   (See http://subversion.tigris.org/faq.html#log-in-source)
--
--
-- Revision 1.15  2005/12/30 14:13:39  jsedwards
-- Removed Social Security Number stuff, not ready yet.
--
-- Revision 1.14  2005/12/30 05:12:57  jsedwards
-- Tweaked the birth date stuff and added social security number.
--
-- Revision 1.13  2005/12/29 18:40:38  jsedwards
-- Added code for the new long key and encryption stuff.
--
-- Revision 1.12  2005/12/27 19:52:21  jsedwards
-- Added code to generate the root object from a password.  Changed to find
-- class definition instead of hardcoded file name and to generate a random
-- person object id instead of generating it based on contents.
--
-- Revision 1.11  2005/12/24 16:18:26  jsedwards
-- Removed "host" id from object references (ObjRef).  Host redirection will
-- be done using a "redirection" object in the future.
--
-- Revision 1.10  2005/12/21 23:28:15  jsedwards
-- Fixed so that it adds the references to the name and date objects.
--
-- Revision 1.9  2005/12/10 15:03:36  jsedwards
-- Fixed header to say the GPL is in the LICENSE file instead of COPYING.
--
-- Revision 1.8  2005/12/05 19:06:31  jsedwards
-- Moved calculation of header checksum down just before writing object to
-- disk to ensure no changes were made after the checksum was calculated.
-- Also added the "nickname" field.
--
-- Revision 1.7  2005/12/04 14:14:46  jsedwards
-- Changed to actually get input and create a person (finally)!
--
-- Revision 1.6  2005/12/04 04:13:02  jsedwards
-- Added 'nwos' prefix to create_xxxx function names and eliminated the
-- 'referring' object parameter from all of them.
--
-- Revision 1.5  2005/12/04 00:35:49  jsedwards
-- Changed to just create a date.
--
-- Revision 1.4  2005/12/04 00:23:46  jsedwards
-- Removed create year, date, and name stuff and put it in the date.c and
-- the name.c files.
--
-- Revision 1.3  2005/12/02 20:33:11  jsedwards
-- Changed so that name objects have "NAME" in the filename and added new
-- parameter to the create_reference_list calls.
--
-- Revision 1.2  2005/12/02 19:30:32  jsedwards
-- Changed to create Year object filename from the year itself.
--
-- Revision 1.1  2005/12/02 13:05:17  jsedwards
-- Initial version that only creates a name and a year.
--
*/

#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "objectify.h"
#include "crc32.h"

#define MAX_NAME_SIZE 64
#define MAX_DATE_SIZE 16
#define MAX_SSN_SIZE  12
#define MAX_WORD_SIZE 12


bool nwos_is_valid_gender(char* str)
{
    if (strcasecmp(str, "female") == 0) return true;
    if (strcasecmp(str, "male") == 0) return true;
    if (strlen(str) == 1)
    {
	if (*str == 'F' || *str == 'f') return true;
	if (*str == 'm' || *str == 'M') return true;
    }

    return false;
}


void nwos_find_public_gender(char* gender, ObjRef* ref)
{
    C_struct_Word word_obj;
    ObjRef gender_class_ref;
    ObjRef english_lang_ref;
    ObjRef word_ref;

    assert(nwos_is_valid_gender(gender));

    assert(nwos_find_public_class_definition("GENDER", &gender_class_ref));

    nwos_find_language("English", &english_lang_ref);

    if (toupper(*gender) == 'F')
    {
	assert(nwos_find_word("female", &gender_class_ref, &english_lang_ref, &word_ref));
    }
    else if (tolower(*gender) == 'm')
    {
	assert(nwos_find_word("male", &gender_class_ref, &english_lang_ref, &word_ref));
    }

    assert(nwos_read_object_from_disk(&word_ref, &word_obj, sizeof(word_obj)));

    memcpy(ref, &word_obj.thing, sizeof(ObjRef));
}


bool nwos_find_private_gender(char* gender, ObjRef* ref)
{
    C_struct_Class_Definition class_def_obj;
    C_struct_Gender gender_obj;
    ObjRef gender_class_ref;
    ObjRef public_gender_ref;
    ObjRef class_ref;
    ReferenceList* ref_list;
    int num_refs;
    int i;
    bool result = false;

    assert(nwos_is_valid_gender(gender));

    nwos_find_public_gender(gender, &public_gender_ref);

    if (nwos_find_private_class_definition("GENDER", &gender_class_ref))
    {
	nwos_read_class_definition(&gender_class_ref, &class_def_obj);

	ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

	num_refs = ref_list->common_header.num_refs;

	for (i = 0; i < num_refs; i++)
	{
	    nwos_get_object_class(&ref_list->references[i], &class_ref);

	    if (is_same_object(&class_ref, &gender_class_ref))
	    {
		assert(nwos_read_object_from_disk(&ref_list->references[i], &gender_obj, sizeof(gender_obj)));

		if (is_same_object(&gender_obj.header.object.clone_of, &public_gender_ref))
		{
		    memcpy(ref, &ref_list->references[i], sizeof(ObjRef));
		    result = true;
		    break;
		}
	    }
	}

	nwos_free_reference_list(ref_list);
	ref_list = NULL;
    }

    return result;
}


/****************************************************************************************/
/* This routine finds a private gender object or creates one by cloning the public one. */
/**************************************(*************************************************/

void nwos_find_or_create_private_gender(char* gender, ObjRef* ref)
{
    ObjRef public_gender_ref;
    ObjRef private_gender_class_ref;
    C_struct_Gender gender_obj;
    char msg[128];

    assert(nwos_is_valid_gender(gender));

    if (!nwos_find_private_gender(gender, ref))
    {
	nwos_find_public_gender(gender, &public_gender_ref);

	nwos_find_or_create_private_class_definition("GENDER", &private_gender_class_ref);

	assert(nwos_read_object_from_disk(&public_gender_ref, &gender_obj, sizeof(gender_obj)));

	nwos_generate_new_id(ref);

	snprintf(msg, sizeof(msg), "clone_gender(%02x%02x%02x%02x) -> %02x%02x%02x%02x\n",
		 public_gender_ref.id[0], public_gender_ref.id[1], public_gender_ref.id[2], public_gender_ref.id[3],
		 ref->id[0], ref->id[1], ref->id[2], ref->id[3]);
	nwos_log(msg);

	copy_reference(&gender_obj.header.common.id, ref);

	nwos_get_time_stamp(gender_obj.header.common.creation_time);

	copy_reference(&gender_obj.header.common.class_definition, &private_gender_class_ref);

	copy_reference(&gender_obj.header.object.clone_of, &public_gender_ref);

	nwos_create_reference_list(ref, &gender_obj.header.object.references);

	nwos_crc32_calculate((uint8*) &gender_obj.header.object, sizeof(ObjectHeader), gender_obj.header.common.header_chksum);

	nwos_write_object_to_disk(ref, &gender_obj, sizeof(gender_obj));

	nwos_add_to_references(ref, &private_gender_class_ref);
    }
}


bool nwos_is_valid_ssn(const char* ssn)
{
    const char *ptr;
    bool result;
    bool dash = false;

    ptr = ssn;

    result = (isdigit(*ptr++) && isdigit(*ptr++) && isdigit(*ptr++));

    if (result && *ptr == '-')
    {
	ptr++;
	dash = true;
    }

    if (result) result = (isdigit(*ptr++) && isdigit(*ptr++));

    if (result && dash) result = (*ptr++ == '-');

    if (result) result = (isdigit(*ptr++) && isdigit(*ptr++) && isdigit(*ptr++) && isdigit(*ptr++));

    return result;
}

#if 0
ObjCreateResult nwos_create_ssn(ObjRef* person, char* number, ObjRef* ref)
{
    C_struct_Class_Definition class_def_obj;
    C_struct_Social_Security_Number ssn;
    ObjRef reference;
    ObjRef class_ref;
    ReferenceList* ref_list;
    int num_ssns;
    char just_digits[9];
    ObjCreateResult result = CREATED_NEW;


    /* first find out if we already have this number */

    assert(nwos_find_private_class_definition("SOCIAL SECURITY NUMBER", &reference));

    nwos_read_class_definition(&reference, &class_def_obj);

    ref_list = nwos_malloc_reference_list(&class_def_obj.header.object.references);

    num_years = ref_list->common_header.num_refs;

    /* printf("num_years: %d\n", num_years); */

    for (i = 0; i < num_years; i++)
    {
	assert(nwos_read_object_from_disk(&ref_list->references[i], kludge, sizeof(kludge)));

	/* remember the ptr_year_obj points to the kludge buffer */

	if (ptr_year_obj->count == 4 && memcmp(ptr_year_obj->storage, year_ascii, 4) == 0)   /* found a match */
	{
	    memcpy(ref, &ref_list->references[i], sizeof(ObjRef));
	    result = FOUND_EXISTING;
	    break;
	}
    }

    nwos_free_reference_list(ref_list);
    ref_list = NULL;

    if (result != FOUND_EXISTING)   /* didn't find it */
    {
	memset(kludge, 0, sizeof(kludge));  /* zero it out */

	/* remember the ptr_year_obj points to the kludge buffer */

	nwos_generate_new_id(ref);

	nwos_fill_in_common_header(&ptr_year_obj->header.common, ref, &reference);

	ptr_year_obj->count = 4;
	memcpy(ptr_year_obj->storage, year_ascii, 4);

	nwos_create_reference_list(ref, &ptr_year_obj->header.object.references);

	nwos_crc32_calculate((uint8*) &ptr_year_obj->header.object, sizeof(ObjectHeader), ptr_year_obj->header.common.header_chksum);

	nwos_crc32_calculate((uint8*) &ptr_year_obj->count, sizeof(kludge) - sizeof(EveryObject), ptr_year_obj->header.common.data_chksum);

	nwos_write_object_to_disk(ref, kludge, sizeof(kludge));

        nwos_add_to_reference_list(ref, &class_def_obj.header.object.references);
    }

    return result;
}
#endif


static void check_public(char* name)
{
    ObjRef ref;

    if (nwos_find_public_name(name, &ref))
    {
	printf("public: %02x%02x%02x%02x\n", ref.id[0], ref.id[1], ref.id[2], ref.id[3]); 
    }
    else
    {
	printf("Warning: %s not in public names\n", name);
    }
}

void nwos_add_person()
{
    ObjRef reference;
    char gender[MAX_WORD_SIZE];
    char first_name[MAX_NAME_SIZE];
    char middle_name[MAX_NAME_SIZE];
    char last_name[MAX_NAME_SIZE];
    char goes_by[MAX_NAME_SIZE];
    char maiden_name[MAX_NAME_SIZE];
    char nickname[MAX_NAME_SIZE];
    char birth_date[MAX_DATE_SIZE];
    char death_date[MAX_DATE_SIZE];
    char temp[MAX_NAME_SIZE];
    ObjRef person_class_ref;
    ObjRef last_name_ref;
    ObjRef middle_name_ref;
    ObjRef first_name_ref;
    ObjRef object_class;
    EveryObject name_header;
    ReferenceList* ref_list;
    int born_month, born_day, born_year;
    int died_month, died_day, died_year;
    int num_refs;
    int i;
    bool ok;
    bool printed_warning = false;
    C_struct_Person person_obj;

#ifdef PUBLIC_MODE
    nwos_find_public_class_definition("PERSON", &person_class_ref);
#else
    nwos_find_or_create_private_class_definition("PERSON", &person_class_ref);
#endif

    ok = false;
    while (!ok)
    {
	nwos_ask_user("gender", gender, sizeof(gender));
	ok = nwos_is_valid_gender(gender);
	if (!ok) printf("must input either 'male' or 'female' or simply 'm' or 'f'\n");
    }

    nwos_ask_user("last name", last_name, sizeof(last_name));
    if (*last_name == '\0')
    {
	printf("Error: this version cannot create a person without the last name.\n");
	return;
    }
    check_public(last_name);

    nwos_ask_user("first name", first_name, sizeof(first_name));
    if (*first_name == '\0')
    {
	printf("Error: this version cannot create a person without the first name.\n");
	return;
    }
    check_public(first_name);

    nwos_ask_user("middle name", middle_name, sizeof(middle_name));
    if (*middle_name != '\0')
    {
	check_public(middle_name);
    }

    nwos_ask_user("goes by", goes_by, sizeof(goes_by));
    if (*goes_by != '\0')
    {
	check_public(goes_by);
    }

    nwos_ask_user("nick name", nickname, sizeof(nickname));
    if (*nickname != '\0')
    {
	check_public(nickname);
    }


    if (toupper(*gender) == 'F')
    {
	nwos_ask_user("maiden name", maiden_name, sizeof(maiden_name));
	if (*maiden_name != '\0')
	{
	    check_public(maiden_name);
	}
    }
    else
    {
	maiden_name[0] = '\0';
    }

    ok = false;
    while (!ok)
    {
	nwos_ask_user("birth date M/D/YYYY", birth_date, sizeof(birth_date));
	if (*birth_date == '\0')
	{
	    ok = true;    /* none given */
	}
	else
	{
	    ok = (sscanf(birth_date, "%d/%d/%d", &born_month, &born_day, &born_year) == 3);
	    if (!ok) ok = (sscanf(birth_date, "%d-%d-%d", &born_month, &born_day, &born_year) == 3);
	    if (ok) ok = (1582 <= born_year && born_year <= 9999);    /* 1582 start of Gregorian calendar */
	    if (ok) ok = (1 <= born_month && born_month <= 12);
	    if (ok) switch (born_month)
	    {
	      case 2:
		if (nwos_is_leap_year(born_year))
		{
		    ok = (born_day <= 29);
		}
		else
		{
		    ok = (born_day <= 28);
		}
		break;

	      case 4:
	      case 6:
	      case 9:
	      case 11:
		ok = (born_day <= 30);
		break;

	      default:
		ok = (born_day <= 31);
		break;
	    }
	}
    }

    ok = false;
    while (!ok)
    {
	nwos_ask_user("death date M/D/YYYY", death_date, sizeof(death_date));
	if (*death_date == '\0')
	{
	    ok = true;    /* none given */
	}
	else
	{
	    ok = (sscanf(death_date, "%d/%d/%d", &died_month, &died_day, &died_year) == 3);
	    if (!ok) ok = (sscanf(death_date, "%d-%d-%d", &died_month, &died_day, &died_year) == 3);
	    if (ok) ok = (1582 <= died_year && died_year <= 9999);    /* 1582 start of Gregorian calendar */
	    if (ok) ok = (1 <= died_month && died_month <= 12);
	    if (ok) switch (died_month)
	    {
	      case 2:
		if (nwos_is_leap_year(died_year))
		{
		    ok = (died_day <= 29);
		}
		else
		{
		    ok = (died_day <= 28);
		}
		break;

	      case 4:
	      case 6:
	      case 9:
	      case 11:
		ok = (died_day <= 30);
		break;

	      default:
		ok = (died_day <= 31);
		break;
	    }
	}
    }

    if (nwos_find_name(last_name, &last_name_ref) && nwos_find_name(first_name, &first_name_ref))
    {
	if (*middle_name == '\0' || !nwos_find_name(middle_name, &middle_name_ref))
	{
	    void_reference(&middle_name_ref);
	}

	nwos_read_object_headers_from_disk(&last_name_ref, &name_header);

	ref_list = nwos_malloc_reference_list(&name_header.object.references);

	num_refs = ref_list->common_header.num_refs;

	for (i = 0; i < num_refs; i++)
	{
	    nwos_get_object_class(&ref_list->references[i], &object_class);   /* find out what kind of object it is */

	    if (is_same_object(&object_class, &person_class_ref))
	    {
		assert(nwos_read_object_from_disk(&ref_list->references[i], &person_obj, sizeof(person_obj)));

		if (is_same_object(&last_name_ref, &person_obj.last_name) && is_same_object(&first_name_ref, &person_obj.first_name))
		{
		    if (*middle_name == '\0' || (!is_void_reference(&middle_name_ref) && is_same_object(&middle_name_ref, &person_obj.middle_name)))
		    {
			if (!printed_warning)
			{
			    printf("\n");
			    printed_warning = true;
			}

			printf("WARNING: a person named ");

			nwos_name_to_string(&person_obj.first_name, temp, sizeof(temp));
			printf("%s ", temp);

			if (!is_void_reference(&person_obj.middle_name))
			{
			    nwos_name_to_string(&person_obj.middle_name, temp, sizeof(temp));
			    printf("%s ", temp);
			}

			nwos_name_to_string(&person_obj.last_name, temp, sizeof(temp));
			printf("%s already exists!\n", temp);
		    }
		}
	    }
	}

	nwos_free_reference_list(ref_list);
	ref_list = NULL;
    }

    printf("\n");

    if (nwos_ask_yes_or_no(NULL, "Create person"))
    {
	memset(&person_obj, 0, sizeof(person_obj));

	nwos_generate_new_id(&reference);

	nwos_fill_in_common_header(&person_obj.header.common, &reference, &person_class_ref);

#ifdef PUBLIC_MODE
	nwos_find_public_gender(gender, &person_obj.gender);
#else
	nwos_find_or_create_private_gender(gender, &person_obj.gender);
#endif

	nwos_create_name(first_name, &person_obj.first_name);

	if (*middle_name != '\0')
	{
	    nwos_create_name(middle_name, &person_obj.middle_name);
	}

	nwos_create_name(last_name, &person_obj.last_name);

	if (*goes_by != '\0') 
	{
	    nwos_create_name(goes_by, &person_obj.goes_by);
	}

	if (*nickname != '\0') 
	{
	    nwos_create_name(nickname, &person_obj.nickname);
	}

	if (*maiden_name != '\0')
	{
	    nwos_create_name(maiden_name, &person_obj.maiden_name);
	}

	if (*birth_date != '\0')
	{
	    nwos_create_date(born_year, born_month, born_day, &person_obj.birth_date);
	}

	if (*death_date != '\0')
	{
	    nwos_create_date(died_year, died_month, died_day, &person_obj.death_date);
	}

	nwos_create_reference_list(&reference, &person_obj.header.object.references);

	nwos_crc32_calculate((uint8*) &person_obj.header.object, sizeof(ObjectHeader), person_obj.header.common.header_chksum);

	nwos_crc32_calculate((uint8*) &person_obj.gender, sizeof(person_obj) - sizeof(EveryObject), person_obj.header.common.data_chksum);

	nwos_write_object_to_disk(&reference, &person_obj, sizeof(person_obj));

        nwos_add_to_references(&reference, &person_class_ref);
	nwos_add_to_references(&reference, &person_obj.gender);
	nwos_add_to_references(&reference, &person_obj.first_name);
	nwos_add_to_references(&reference, &person_obj.last_name);

	if (!is_void_reference(&person_obj.middle_name))
	{
	    nwos_add_to_references(&reference, &person_obj.middle_name);
	}

	if (!is_void_reference(&person_obj.goes_by))
	{
	    nwos_add_to_references(&reference, &person_obj.goes_by);
	}

	if (!is_void_reference(&person_obj.nickname))
	{
	    nwos_add_to_references(&reference, &person_obj.nickname);
	}

	if (!is_void_reference(&person_obj.maiden_name))
	{
	    nwos_add_to_references(&reference, &person_obj.maiden_name);
	}

	if (!is_void_reference(&person_obj.birth_date))
	{
	    nwos_add_to_references(&reference, &person_obj.birth_date);
	}
	
	if (!is_void_reference(&person_obj.death_date))
	{
	    nwos_add_to_references(&reference, &person_obj.death_date);
	}
	
	printf("Object created: %02x%02x%02x%02x\n", reference.id[0], reference.id[1], reference.id[2], reference.id[3]);
    }
}


void nwos_add_birthday(char* name, char* date)
{
    bool date_ok;
    ObjRef person_ref;
    int month, day, year;
    C_struct_Person person_obj;
#if 0
    char temp[16];
#endif

    date_ok = (sscanf(date, "%d/%d/%d", &month, &day, &year) == 3);
    if (!date_ok) date_ok = (sscanf(date, "%d-%d-%d", &month, &day, &year) == 3);

    if (!date_ok)
    {
	printf("Date must be of the format MM/DD/YYYY or MM-DD-YYYY\n");
    }
    else
    {
	date_ok = (1582 <= year && year <= 9999);    /* 1582 start of Gregorian calendar */

	if (!date_ok)
	{
	    printf("Year must be in the range 1582 to 9999\n");
	}
	else
	{
	    date_ok = (1 <= month && month <= 12);

	    if (!date_ok)
	    {
		printf("Month must be in the range 1 to 12\n");
	    }
	    else
	    {
		switch (month)
		{
		  case 2:
		    if (nwos_is_leap_year(year))
		    {
			date_ok = (1 <= day && day <= 29);

			if (!date_ok)
			{
			    printf("Day must be in the range 1 to 29 because %d is a leap year\n", year);
			}
		    }
		    else
		    {
			date_ok = (1 <= day && day <= 28);

			if (!date_ok)
			{
			    printf("Day must be in the range 1 to 28 because %d is not a leap year\n", year);
			}
		    }
		    break;

		  case 4:
		  case 6:
		  case 9:
		  case 11:
		    date_ok = (1 <= day && day <= 30);
		    if (!date_ok)
		    {
			printf("Day must be in the range 1 to 30\n");
		    }
		    break;

		  default:
		    date_ok = (1 <= day && day <= 31);
		    if (!date_ok)
		    {
			printf("Day must be in the range 1 to 31\n");
		    }
		    break;
		}
	    }
	}
    }

    if (!date_ok) return;

    if (!nwos_find_person(name, &person_ref))
    {
	printf("I'm sorry I can't find %s in the database.\n", name);
	return;
    }

    assert(nwos_read_object_from_disk(&person_ref, &person_obj, sizeof(person_obj)));

    if (!is_void_reference(&person_obj.birth_date))
    {
	printf("This version cannot change a birthday, SORRY!\n");
	return;
#if 0
	nwos_date_to_string(&person_obj.birth_date, temp, sizeof(temp));

	printf("%s already has a birthday set: %s\n", name, temp);

	if (nwos_ask_yes_or_no(NULL, "overwrite"))
	{
	    /* need to make sure we can delete reference from old date */
	}
#endif
    }  
	
    nwos_create_date(year, month, day, &person_obj.birth_date);

    nwos_crc32_calculate((uint8*) &person_obj.gender, sizeof(person_obj) - sizeof(EveryObject), person_obj.header.common.data_chksum);

    nwos_overwrite_object_to_disk(&person_ref, &person_obj, sizeof(person_obj));

    nwos_add_to_references(&person_ref, &person_obj.birth_date);
}


bool nwos_any_persons_named(char* name)
{
    ObjRef name_ref;
    ObjRef object_class;
    EveryObject name_header;
    ObjRef person_class_ref;
    ReferenceList* ref_list;
    int num_refs;
    int i;
    bool result = false;

  /* first find the name */

    if (nwos_find_private_name(name, &name_ref))
    {
	assert(nwos_find_private_class_definition("PERSON", &person_class_ref));

	nwos_read_object_headers_from_disk(&name_ref, &name_header);

	ref_list = nwos_malloc_reference_list(&name_header.object.references);

	num_refs = ref_list->common_header.num_refs;

	/* printf("name num refs: %d\n", num_refs); */

	for (i = 0; i < num_refs; i++)
	{
	    nwos_get_object_class(&ref_list->references[i], &object_class);   /* find out what kind of object it is */

	    if (is_same_object(&object_class, &person_class_ref))
	    {
		result = true;
		break;
	    }
	}
    }

    return result;
}


#define MAX_NAMES 20

bool nwos_find_person(char* name, ObjRef* person_ref)
{
    C_struct_Person person_obj;
    ObjRef name_ref;
    ObjRef object_class;
    EveryObject name_header;
    ObjRef person_class_ref;
    ObjRef person_refs[MAX_NAMES];
    int person_count;
    ReferenceList* ref_list;
    int num_refs;
    int selection;
    char temp[64];
    char buffer[4];
    bool ok;
    int i;
    int j;
    bool result = false;

  /* first find the name */

    if (nwos_find_private_name(name, &name_ref))
    {
	assert(nwos_find_private_class_definition("PERSON", &person_class_ref));

	nwos_read_object_headers_from_disk(&name_ref, &name_header);

	ref_list = nwos_malloc_reference_list(&name_header.object.references);

	num_refs = ref_list->common_header.num_refs;

	/* printf("name num refs: %d\n", num_refs); */

	person_count = 0;

	for (i = 0; i < num_refs; i++)
	{
	    nwos_get_object_class(&ref_list->references[i], &object_class);   /* find out what kind of object it is */

	    if (is_same_object(&object_class, &person_class_ref))
	    {
		for (j = 0; j < person_count; j++)    /* see if we already have this one */
		{
		    if (is_same_object(&ref_list->references[i], &person_refs[j])) break;
		}

		if (j == person_count)    /* not already in the list, so add it */
		{
		    memcpy(&person_refs[person_count], &ref_list->references[i], sizeof(ObjRef));
		    person_count++;
		}
	    }

	    if (person_count == MAX_NAMES)
	    {
		printf("/nSorry, right now this program can only deal with a maximum of %d possible names.\n", MAX_NAMES);
		break;
	    }
	}

	nwos_free_reference_list(ref_list);
	ref_list = NULL;

	selection = 0;    /* default to first person in case there is only one */

	if (person_count > 1)  /* ask which one we want */
	{
	    printf("\nThere is more than one person with that name:\n\n");

	    for (j = 0; j < person_count; j++)    /* see if we already have this one */
	    {
		assert(nwos_read_object_from_disk(&person_refs[j], &person_obj, sizeof(person_obj)));

		nwos_name_to_string(&person_obj.first_name, temp, sizeof(temp));
		printf("  %d: %s ", j, temp);
		if (!is_void_reference(&person_obj.middle_name))
		{
		    nwos_name_to_string(&person_obj.middle_name, temp, sizeof(temp));
		    printf("%s ", temp);
		}
		nwos_name_to_string(&person_obj.last_name, temp, sizeof(temp));
		printf("%s\n", temp);
	    }

	    printf("Which %s are you referring to?\n\n", name);

	    ok = false;
	    while (!ok)
	    {
		nwos_ask_user("selection number", buffer, sizeof(buffer));

		printf("\n");

		if (isdigit(buffer[0]) && buffer[1] == '\0')   /* single digit */
		{
		    selection = buffer[0] - '0';
		    ok = (selection <= person_count);
		}
		else if (isdigit(buffer[0]) && buffer[2] == '\0')   /* two digits */
		{
		    selection = (buffer[0] - '0') * 10 + (buffer[1] - '0');
		    ok = (selection < person_count);
		}

		if (!ok) printf("Please enter the number of one of the people listed.\n");
	    }
	}

	if (person_count > 0)
	{
	    memcpy(person_ref, &person_refs[selection], sizeof(ObjRef));
	    result = true;
	}
    }

    return result;
}


void nwos_list_persons()
{
    C_struct_Person person_obj;
    ObjRef object_class;
    EveryObject class_ref_header;
    ObjRef person_class_ref;
    ReferenceList* ref_list;
    int num_refs;
    int i;
    char name[32];

#ifdef PUBLIC_MODE
    if (!nwos_find_public_class_definition("PERSON", &person_class_ref))
#else
    if (!nwos_find_private_class_definition("PERSON", &person_class_ref))
#endif
    {
	printf("\nThere are no persons in the system!\n\n");
	return;
    }

    nwos_read_object_headers_from_disk(&person_class_ref, &class_ref_header);

    ref_list = nwos_malloc_reference_list(&class_ref_header.object.references);

    num_refs = ref_list->common_header.num_refs;

    /* printf("person class num refs: %d\n", num_refs); */

    for (i = 0; i < num_refs; i++)
    {
	nwos_get_object_class(&ref_list->references[i], &object_class);   /* find out what kind of object it is */

	if (is_same_object(&object_class, &person_class_ref))
	{
	    assert(nwos_read_object_from_disk(&ref_list->references[i], &person_obj, sizeof(person_obj)));

	    nwos_name_to_string(&person_obj.last_name, name, sizeof(name));

	    printf("%s, ", name);

	    nwos_name_to_string(&person_obj.first_name, name, sizeof(name));

	    printf("%s ", name);

	    if (!is_void_reference(&person_obj.middle_name))
	    {
		nwos_name_to_string(&person_obj.middle_name, name, sizeof(name));

		printf("%s", name);
	    }

	    printf("\n");
	}
    }

    nwos_free_reference_list(ref_list);
    ref_list = NULL;
}


void nwos_describe_person(char* name)
{
    C_struct_Person person_obj;
    ObjRef person_ref;
    char temp[64];
    
    if (!nwos_find_person(name, &person_ref))
    {
	printf("\nI'm sorry I don't know anyone named %s.\n\n", name);
    }
    else
    {
	assert(nwos_read_object_from_disk(&person_ref, &person_obj, sizeof(person_obj)));

	printf("full name: ");

	nwos_name_to_string(&person_obj.first_name, temp, sizeof(temp));
	printf("%s ", temp);

	if (!is_void_reference(&person_obj.middle_name))
	{
	    nwos_name_to_string(&person_obj.middle_name, temp, sizeof(temp));
	    printf("%s ", temp);
	}

	nwos_name_to_string(&person_obj.last_name, temp, sizeof(temp));
	printf("%s\n", temp);


	if (!is_void_reference(&person_obj.goes_by))
	{
	    printf("goes by: ");
	    nwos_name_to_string(&person_obj.goes_by, temp, sizeof(temp));
	    printf("%s\n", temp);
	}

	if (!is_void_reference(&person_obj.nickname))
	{
	    printf("nickname: ");
	    nwos_name_to_string(&person_obj.nickname, temp, sizeof(temp));
	    printf("%s\n", temp);
	}

	if (!is_void_reference(&person_obj.maiden_name))
	{
	    printf("maiden_name: ");
	    nwos_name_to_string(&person_obj.maiden_name, temp, sizeof(temp));
	    printf("%s\n", temp);
	}

	if (!is_void_reference(&person_obj.birth_date))
	{
	    printf("born: ");
	    nwos_date_to_string(&person_obj.birth_date, temp, sizeof(temp));
	    printf("%s\n", temp);
	}

	if (!is_void_reference(&person_obj.death_date))
	{
	    printf("died: ");
	    nwos_date_to_string(&person_obj.death_date, temp, sizeof(temp));
	    printf("%s\n", temp);
	}
    }

    printf("\n");
}


void nwos_full_name_to_string(ObjRef* first_name_ref, ObjRef* middle_name_ref, ObjRef* last_name_ref, char* name, size_t size)
{
    char first_name[32];
    char middle_name[32];
    char last_name[32];

    nwos_name_to_string(first_name_ref, first_name, sizeof(first_name));

    if (is_void_reference(middle_name_ref))
    {
	middle_name[0] = '\0';
    }
    else
    {
	nwos_name_to_string(middle_name_ref, middle_name, sizeof(middle_name));
    }

    nwos_name_to_string(last_name_ref, last_name, sizeof(last_name));

    snprintf(name, size, "%s %s %s", first_name, middle_name, last_name);
}


